# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.ops.environments.resources import PodResourcesSchema


class PodEnvironmentSchema(BaseSchema):
    # To indicate which worker/ps index this session config belongs to
    index = fields.Int(allow_none=True)
    resources = fields.Nested(PodResourcesSchema, allow_none=True)
    node_selector = fields.Dict(allow_none=True)
    affinity = fields.Dict(allow_none=True)
    tolerations = fields.List(fields.Dict(), allow_none=True)

    @staticmethod
    def schema_config():
        return PodEnvironmentConfig


class PodEnvironmentConfig(BaseConfig):
    """
    Pod environment config.

    Args:
        index: `int | None`. The index of the pod.
        resources: `PodResourcesConfig`.
        node_selector: `dict`.
        affinity: `dict`.
        tolerations: `list(dict)`.
    """
    IDENTIFIER = 'pod_environment'
    SCHEMA = PodEnvironmentSchema
    REDUCED_ATTRIBUTES = ['index', 'resources', 'node_selector', 'affinity', 'tolerations']

    def __init__(self,
                 index=None,
                 resources=None,
                 node_selector=None,
                 affinity=None,
                 tolerations=None,
                 ):
        self.index = index
        self.resources = resources
        self.node_selector = node_selector
        self.affinity = affinity
        self.tolerations = tolerations



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseMultiSchema
from polyaxon_schemas.ml.layers.base import BaseLayerConfig, BaseLayerSchema
from polyaxon_schemas.utils import ObjectOrListObject


class AddSchema(BaseLayerSchema):

    @staticmethod
    def schema_config():
        return AddConfig


class AddConfig(BaseLayerConfig):
    """Layer that adds a list of inputs.

    It takes as input a list of tensors,
    all of the same shape, and returns
    a single tensor (also of the same shape).

    Polyaxonfile usage:

    ```yaml
    - Dense:
        units: 10
        activation: softmax
        name: dense1

    - Dense:
        units: 10
        activation: softmax
        name: dense2

    - Merge:
        Add:
          inbound_nodes: [dense1, dense2]
    ```
    """
    IDENTIFIER = 'Add'
    SCHEMA = AddSchema


class SubtractSchema(BaseLayerSchema):

    @staticmethod
    def schema_config():
        return SubtractConfig


class SubtractConfig(BaseLayerConfig):
    """Layer that subtracts two inputs.

    It takes as input a list of tensors of size 2,
    both of the same shape, and returns a single tensor, (inputs[0] - inputs[1]),
    also of the same shape.

    Examples:

    ```python
    # input1.shape == (16,)
    x1 = Dense(8, activation='relu')(input1)
    # input2.shape == (32,)
    x2 = Dense(8, activation='relu')(input2)
    subtracted = Subtract()([x1, x2])
    ```

    Polyaxonfile usage:

    ```yaml
    - Dense:
        units: 10
        activation: softmax
        name: dense1

    - Dense:
        units: 10
        activation: softmax
        name: dense2

    - Merge:
        Subtract:
          inbound_nodes: [dense1, dense2]
    ```
    """
    IDENTIFIER = 'Subtract'
    SCHEMA = SubtractSchema


class MultiplySchema(BaseLayerSchema):

    @staticmethod
    def schema_config():
        return MultiplyConfig


class MultiplyConfig(BaseLayerConfig):
    """Layer that multiplies (element-wise) a list of inputs.

    It takes as input a list of tensors,
    all of the same shape, and returns
    a single tensor (also of the same shape).

    Polyaxonfile usage:

    ```yaml
    - Dense:
        units: 10
        activation: softmax
        name: dense1

    - Dense:
        units: 10
        activation: softmax
        name: dense2

    - Merge:
        Multiply:
          inbound_nodes: [dense1, dense2]
    ```
    """
    IDENTIFIER = 'Multiply'
    SCHEMA = MultiplySchema


class AverageSchema(BaseLayerSchema):

    @staticmethod
    def schema_config():
        return AverageConfig


class AverageConfig(BaseLayerConfig):
    """Layer that averages a list of inputs.

    It takes as input a list of tensors,
    all of the same shape, and returns
    a single tensor (also of the same shape).

    Polyaxonfile usage:

    ```yaml
    - Dense:
        units: 10
        activation: softmax
        name: dense1

    - Dense:
        units: 10
        activation: softmax
        name: dense2

    - Merge:
        Average:
          inbound_nodes: [dense1, dense2]
    ```
    """
    IDENTIFIER = 'Average'
    SCHEMA = AverageSchema


class MaximumSchema(BaseLayerSchema):

    @staticmethod
    def schema_config():
        return MaximumConfig


class MaximumConfig(BaseLayerConfig):
    """Layer that computes the maximum (element-wise) a list of inputs.

    It takes as input a list of tensors,
    all of the same shape, and returns
    a single tensor (also of the same shape).

    Polyaxonfile usage:

    ```yaml
    - Dense:
        units: 10
        activation: softmax
        name: dense1

    - Dense:
        units: 10
        activation: softmax
        name: dense2

    - Merge:
        Maximum:
          inbound_nodes: [dense1, dense2]
    ```
    """
    IDENTIFIER = 'Maximum'
    SCHEMA = MaximumSchema


class ConcatenateSchema(BaseLayerSchema):
    axis = fields.Int(default=-1, missing=-1)

    @staticmethod
    def schema_config():
        return ConcatenateConfig


class ConcatenateConfig(BaseLayerConfig):
    """Layer that concatenates a list of inputs.

    It takes as input a list of tensors,
    all of the same shape expect for the concatenation axis,
    and returns a single tensor, the concatenation of all inputs.

    Args:
        axis: Axis along which to concatenate.
        **kwargs: standard layer keyword arguments.

    Polyaxonfile usage:

    ```yaml
    - Dense:
        units: 10
        activation: softmax
        name: dense1

    - Dense:
        units: 10
        activation: softmax
        name: dense2

    - Merge:
        Concatenate:
          axis: 1
          inbound_nodes: [dense1, dense2]
    ```
    """
    IDENTIFIER = 'Concatenate'
    SCHEMA = ConcatenateSchema

    def __init__(self, axis=-1, **kwargs):
        super(ConcatenateConfig, self).__init__(**kwargs)
        self.axis = axis


class DotSchema(BaseLayerSchema):
    axes = ObjectOrListObject(fields.Int)
    normalize = fields.Bool(allow_none=True)

    @staticmethod
    def schema_config():
        return DotConfig


class DotConfig(BaseLayerConfig):
    """Layer that computes a dot product between samples in two tensors.

    E.g. if applied to two tensors `a` and `b` of shape `(batch_size, n)`,
    the output will be a tensor of shape `(batch_size, 1)`
    where each entry `i` will be the dot product between
    `a[i]` and `b[i]`.

    Args:
        axes: Integer or tuple of integers,
            axis or axes along which to take the dot product.
        normalize: Whether to L2-normalize samples along the
            dot product axis before taking the dot product.
            If set to True, then the output of the dot product
            is the cosine proximity between the two samples.
        **kwargs: Standard layer keyword arguments.

    Polyaxonfile usage:

    ```yaml
    - Dense:
        units: 10
        activation: softmax
        name: dense1

    - Dense:
        units: 10
        activation: softmax
        name: dense2

    - Merge:
        Dot:
          axes: [1, 2]
          inbound_nodes: [dense1, dense2]
    ```
    """
    IDENTIFIER = 'Dot'
    SCHEMA = DotSchema

    def __init__(self, axes, normalize=False, **kwargs):
        super(DotConfig, self).__init__(**kwargs)
        self.axes = axes
        self.normalize = normalize


class MergeSchema(BaseMultiSchema):
    __multi_schema_name__ = 'Merge'
    __configs__ = {
        AddConfig.IDENTIFIER: AddConfig,
        MultiplyConfig.IDENTIFIER: MultiplyConfig,
        AverageConfig.IDENTIFIER: AverageConfig,
        MaximumConfig.IDENTIFIER: MaximumConfig,
        ConcatenateConfig.IDENTIFIER: ConcatenateConfig,
        DotConfig.IDENTIFIER: DotConfig,
    }


class MergeConfig(BaseLayerConfig):
    IDENTIFIER = 'Merge'
    SCHEMA = MergeSchema



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import os

from unittest import TestCase

from marshmallow import ValidationError
from tests.utils import assert_equal_dict

from polyaxon_schemas.exceptions import PolyaxonConfigurationError, PolyaxonfileError
from polyaxon_schemas.ops.environments.experiments import ExperimentEnvironmentConfig
from polyaxon_schemas.ops.environments.resources import K8SResourcesConfig, PodResourcesConfig
from polyaxon_schemas.ops.experiment import ExperimentConfig
from polyaxon_schemas.specs import (
    BuildSpecification,
    ExperimentSpecification,
    GroupSpecification,
    JobSpecification,
    NotebookSpecification,
    TensorboardSpecification
)
from polyaxon_schemas.utils import TaskType


class TestSpecifications(TestCase):
    def test_notebook_specification_raises_for_invalid_run_section(self):
        with self.assertRaises(PolyaxonfileError):
            NotebookSpecification.read({'version': 1, 'kind': 'notebook'})

        with self.assertRaises(PolyaxonfileError):
            NotebookSpecification.read(os.path.abspath(
                'tests/fixtures/notebook_run_exec_simple_file_with_cmd.yml'))

    def test_tensorboard_specification_raises_for_invalid_run_section(self):
        with self.assertRaises(PolyaxonfileError):
            TensorboardSpecification.read({'version': 1, 'kind': 'tensorboard'})

        with self.assertRaises(PolyaxonConfigurationError):
            TensorboardSpecification.read(os.path.abspath(
                'tests/fixtures/tensorboard_run_exec_simple_file_with_cmd.yml'))

    def test_job_specification_raises_for_missing_build_section(self):
        with self.assertRaises(PolyaxonfileError):
            JobSpecification.read(os.path.abspath(
                'tests/fixtures/job_missing_build.yml'))

    def test_job_specification_raises_for_missing_run_section(self):
        with self.assertRaises(PolyaxonfileError):
            JobSpecification.read(os.path.abspath(
                'tests/fixtures/job_missing_run_exec.yml'))

    def test_create_notebook_specification(self):
        build_config = {'image': 'blabla'}
        config = NotebookSpecification.create_specification(build_config)
        assert NotebookSpecification.read(config).parsed_data == config
        assert config['build'] == build_config
        spec = NotebookSpecification.create_specification(build_config, to_dict=False)
        assert spec.build.image == build_config['image']

    def test_create_tensorboard_specification(self):
        build_config = {'image': 'blabla'}
        config = TensorboardSpecification.create_specification(build_config)
        assert TensorboardSpecification.read(config).parsed_data == config
        assert config['build'] == build_config
        spec = TensorboardSpecification.create_specification(build_config, to_dict=False)
        assert spec.build.image == build_config['image']

    def test_create_build_specification(self):
        # Normal build config
        build_config = {'image': 'blabla'}
        config = BuildSpecification.create_specification(build_config)
        assert BuildSpecification.read(config).parsed_data == config
        assert config['image'] == build_config['image']
        spec = BuildSpecification.create_specification(build_config, to_dict=False)
        assert spec.config.image == build_config['image']

        # Run config
        run_config = {'image': 'blabla', 'cmd': 'some command'}
        config = BuildSpecification.create_specification(run_config)
        assert BuildSpecification.read(config).parsed_data == config

        config = BuildSpecification.create_specification(run_config,
                                                         configmap_refs=None,
                                                         secret_refs=None)
        assert 'configmap_refs' not in config
        assert 'secret_refs' not in config

        config = BuildSpecification.create_specification(run_config, configmap_refs=['foo'])
        assert BuildSpecification.read(config).parsed_data == config
        assert config['environment']['configmap_refs'] == ['foo']
        config = BuildSpecification.create_specification(run_config, secret_refs=['foo'])
        assert BuildSpecification.read(config).parsed_data == config
        assert config['environment']['secret_refs'] == ['foo']

        assert config['image'] == 'blabla'
        spec = BuildSpecification.create_specification(run_config, to_dict=False)
        assert spec.config.image == run_config['image']

        assert config['image'] == 'blabla'
        spec = BuildSpecification.create_specification(run_config,
                                                       configmap_refs=['foo'],
                                                       secret_refs=['foo'],
                                                       to_dict=False)
        assert spec.config.image == run_config['image']
        assert spec.environment.secret_refs == ['foo']
        assert spec.environment.configmap_refs == ['foo']

    def test_create_job_specification(self):
        build_config = {'image': 'blabla'}
        run_config = {'cmd': 'some command'}
        config = JobSpecification.create_specification(build_config=build_config,
                                                       run_config=run_config)
        assert JobSpecification.read(config).parsed_data == config
        assert config['run'] == run_config
        assert config['build'] == build_config
        spec = JobSpecification.create_specification(build_config=build_config,
                                                     run_config=run_config,
                                                     to_dict=False)
        assert spec.build.image == build_config['image']
        assert spec.run.cmd == run_config['cmd']

    def test_cluster_def_without_framework(self):
        spec = ExperimentSpecification.read(os.path.abspath(
            'tests/fixtures/env_without_framework.yml'))
        self.assertEqual(spec.cluster_def, ({TaskType.MASTER: 1}, False))

    def test_patch_experiment(self):
        content = {
            'version': 1,
            'kind': 'experiment',
            'build': {'image': 'my_image'},
            'run': {'cmd': 'train'}
        }
        spec = ExperimentSpecification.read(content)
        assert spec.declarations is None

        # Add declarations
        declarations = {'declarations': {'lr': 0.1}}
        spec = spec.patch(values=declarations)
        assert spec.declarations == declarations['declarations']

        # Update declarations
        declarations = {'declarations': {'lr': 0.01, 'num_steps': 100}}
        spec = spec.patch(values=declarations)
        assert spec.declarations == declarations['declarations']

        # Add env
        assert spec.environment is None
        env = {'environment': {'resources': {
            'gpu': {'requests': 1, 'limits': 1},
            'tpu': {'requests': 1, 'limits': 1}}}}
        spec = spec.patch(values=env)
        assert spec.declarations == declarations['declarations']
        assert spec.environment.resources.gpu.to_dict() == env['environment']['resources']['gpu']
        assert spec.environment.resources.tpu.to_dict() == env['environment']['resources']['gpu']

        # Patch with unsupported spec
        matrix = {'hptuning': {'matrix': {'lr': {'values': [0.1, 0.2]}}}}
        with self.assertRaises(PolyaxonfileError):
            spec.patch(values=matrix)

        # Patch with unsupported spec
        wrong_config = {'lr': {'values': [0.1, 0.2]}}
        with self.assertRaises(PolyaxonfileError):
            spec.patch(values=wrong_config)

    def test_experiment_environment_config(self):
        config_dict = {
            'resources': PodResourcesConfig(cpu=K8SResourcesConfig(0.5, 1)).to_dict(),
            'replicas': {
                'n_workers': 10,
                'n_ps': 5,
            }
        }
        config = ExperimentEnvironmentConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add some field should raise
        config_dict['foo'] = {
            'n_workers': 10,
            'n_ps': 5,
        }

        with self.assertRaises(ValidationError):
            ExperimentEnvironmentConfig.from_dict(config_dict)

        del config_dict['foo']

        experiment_config = {
            'environment': config_dict,
            'framework': 'tensorflow'
        }
        config = ExperimentConfig.from_dict(experiment_config)
        assert_equal_dict(experiment_config, config.to_dict())

        # Removing framework tensorflow should raise
        del experiment_config['framework']
        with self.assertRaises(ValidationError):
            ExperimentConfig.from_dict(experiment_config)

        # Using unknown framework should raise
        experiment_config['framework'] = 'foo'
        with self.assertRaises(ValidationError):
            ExperimentConfig.from_dict(experiment_config)

        # Using known framework
        experiment_config['framework'] = 'mxnet'
        config = ExperimentConfig.from_dict(experiment_config)
        assert_equal_dict(experiment_config, config.to_dict())

        # Adding horovod should raise
        experiment_config['framework'] = 'horovod'
        with self.assertRaises(ValidationError):
            ExperimentConfig.from_dict(experiment_config)

        # Setting correct horovod replicas should pass
        experiment_config['environment']['replicas'] = {
            'n_workers': 5
        }
        config = ExperimentConfig.from_dict(experiment_config)
        assert_equal_dict(experiment_config, config.to_dict())

        # Adding pytorch should pass
        experiment_config['framework'] = 'pytorch'
        config = ExperimentConfig.from_dict(experiment_config)
        assert_equal_dict(experiment_config, config.to_dict())

        # Setting wrong pytorch replicas should raise
        experiment_config['environment']['replicas'] = {
            'n_workers': 5,
            'n_ps': 1
        }

        with self.assertRaises(ValidationError):
            ExperimentConfig.from_dict(experiment_config)

    def test_group_environment(self):
        content = {
            'version': 1,
            'kind': 'group',
            'hptuning': {'matrix': {'lr': {'values': [0.1, 0.2]}}},
            'build': {'image': 'my_image'},
            'run': {'cmd': 'train'}
        }
        spec = GroupSpecification.read(content)
        assert spec.environment is None
        assert spec.configmap_refs is None
        assert spec.secret_refs is None

        content['environment'] = {'configmap_refs': ['foo', 'boo']}
        spec = GroupSpecification.read(content)
        assert spec.environment is not None
        assert spec.configmap_refs == ['foo', 'boo']
        assert spec.secret_refs is None

        content['environment'] = {'secret_refs': ['foo', 'boo']}
        spec = GroupSpecification.read(content)
        assert spec.environment is not None
        assert spec.configmap_refs is None
        assert spec.secret_refs == ['foo', 'boo']

        content['environment'] = {'secret_refs': ['foo', 'boo'], 'configmap_refs': ['foo', 'boo']}
        spec = GroupSpecification.read(content)
        assert spec.environment is not None
        assert spec.configmap_refs == ['foo', 'boo']
        assert spec.secret_refs == ['foo', 'boo']



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_dict

from polyaxon_schemas.ml.hooks import (
    EpisodeLoggingTensorHookConfig,
    FinalOpsHookConfig,
    GlobalStepWaiterHookConfig,
    StepLoggingTensorHookConfig
)


class TestHookConfigs(TestCase):
    def test_global_step_wait_hook_config(self):
        config_dict = {
            'wait_until_step': 10
        }
        config = GlobalStepWaiterHookConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert_equal_dict(config_to_dict, config_dict)

    def test_final_ops_hook_config(self):
        config_dict = {
            'final_ops': ['loss', 'precision']
        }
        config = FinalOpsHookConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert_equal_dict(config_to_dict, config_dict)

    def test_step_logging_tensor_hook(self):
        config_dict = {
            'tensors': ['conv2d_1', 'relu_1'],
            'every_n_iter': 10,
            'every_n_secs': None
        }
        config = StepLoggingTensorHookConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert_equal_dict(config_to_dict, config_dict)

    def test_episode_logging_tensor_hook(self):
        config_dict = {
            'tensors': ['conv2d_1', 'relu_1'],
            'every_n_episodes': 2,
        }
        config = EpisodeLoggingTensorHookConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert_equal_dict(config_to_dict, config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.ops.build import BuildSchema
from polyaxon_schemas.ops.environments.base import EnvironmentSchema
from polyaxon_schemas.ops.logging import LoggingSchema


class BaseRunSchema(BaseSchema):
    version = fields.Int(allow_none=None)
    kind = fields.Str(allow_none=None, validate=validate.Equal('Implement'))
    logging = fields.Nested(LoggingSchema, allow_none=None)
    tags = fields.List(fields.Str(), allow_none=None)
    environment = fields.Nested(EnvironmentSchema, allow_none=True)
    build = fields.Nested(BuildSchema)

    @staticmethod
    def schema_config():
        return BaseRunConfig


class BaseRunConfig(BaseConfig):
    """
    Build config.

    Args:
        image: str. The name of the image to use during the build step.
        build_steps: list(str). The build steps to apply to your docker image.
            (translate to multiple RUN ...)
        env_vars: list((str, str)) The environment variable to set on you docker image.
        nocache: `bool`. To not use cache when building the image.
        ref: `str`. The commit/branch/treeish to use.

    """
    SCHEMA = BaseRunSchema
    IDENTIFIER = 'run'
    REDUCED_ATTRIBUTES = [
        'kind',
        'version',
        'logging',
        'tags',
        'environment',
    ]

    def __init__(self,
                 kind=None,
                 version=None,
                 logging=None,
                 tags=None,
                 environment=None,
                 build=None,
                 ):
        self.kind = kind
        self.version = version
        self.logging = logging
        self.tags = tags
        self.environment = environment
        self.build = build



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from marshmallow import ValidationError
from tests.utils import assert_equal_dict

from polyaxon_schemas.ops.environments.resources import K8SResourcesConfig, PodResourcesConfig
from polyaxon_schemas.ops.notebook import NotebookConfig


class TestNotebookEnvironmentsConfigs(TestCase):

    def test_notebook_environment_config(self):
        config_dict = {
            'environment': {
                'resources': PodResourcesConfig(cpu=K8SResourcesConfig(0.5, 1)).to_dict(),
            }
        }
        config = NotebookConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        config_dict['backend'] = 'notebook'
        config = NotebookConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        config_dict['backend'] = 'lab'
        config = NotebookConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        config_dict['backend'] = 'foo'
        with self.assertRaises(ValidationError):
            NotebookConfig.from_dict(config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.utils import UUID


class NodeGPUSchema(BaseSchema):
    index = fields.Int()
    name = fields.Str()
    uuid = UUID()
    memory = fields.Int()
    serial = fields.Str()
    cluster_node = UUID()

    @staticmethod
    def schema_config():
        return NodeGPUConfig


class NodeGPUConfig(BaseConfig):
    """
    Node gpu config.

    Args:
        index: `int`. The index of the gpu during the discovery.
        name: `str`. The name of gpu.
        uuid: `UUID`. The uuid of gpu.
        memory: `int`. The memory size of the gpu.
        serial: `str`. The serial of the gpu.
        cluster_node: `UUID`. the uuid of the cluster node.
    """
    SCHEMA = NodeGPUSchema
    IDENTIFIER = 'NodeGPU'
    DEFAULT_EXCLUDE_ATTRIBUTES = ['uuid', 'cluster_node']

    def __init__(self, index, name, uuid, memory, serial, cluster_node):
        self.uuid = uuid
        self.serial = serial
        self.name = name
        self.index = index
        self.memory = memory
        self.cluster_node = cluster_node


class ClusterNodeSchema(BaseSchema):
    sequence = fields.Int(allow_none=True)
    name = fields.Str(allow_none=True)
    uuid = UUID()
    status = fields.Str(allow_none=True)
    hostname = fields.Str(allow_none=True)
    role = fields.Str(allow_none=True)
    memory = fields.Int(allow_none=True)
    cpu = fields.Float(allow_none=True)
    n_gpus = fields.Int(allow_none=True)
    kubelet_version = fields.Str(allow_none=True)
    docker_version = fields.Str(allow_none=True)
    os_image = fields.Str(allow_none=True)
    kernel_version = fields.Str(allow_none=True)
    schedulable_taints = fields.Bool(allow_none=True)
    schedulable_state = fields.Bool(allow_none=True)
    gpus = fields.Nested(NodeGPUSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return ClusterNodeConfig


class ClusterNodeConfig(BaseConfig):
    """
    Node gpu config.

    Args:
        uuid: `UUID`. the uuid of the cluster node.
        sequence: `int`. The sequence of the node in the cluster.
        name: `str`. The name of node.
        hostname: `str`. The node hostname.
        role: `str`. The role of the node.
        docker_version: `str`. The docker version used in the node.
        kubelet_version: `str`. The kubelet version used in the node.
        os_image: `str`. The os image used of the node.
        kernel_version: `str`. The kernel version of the node.
        schedulable_taints: `bool`. The schedulable taints of the node.
        schedulable_state: `bool`. The schedulable state of the node.
        memory: `int`. The memory size of the node.
        cpu: `float`. The cpu of the node.
        n_gpus: `int`. The number of gpus in the node.
        status: `str`. The status of the node (ready or ...)
        gpus: `list(NodeGPUConfig)`. The node gpus.
    """
    SCHEMA = ClusterNodeSchema
    IDENTIFIER = 'ClusterNode'
    DEFAULT_INCLUDE_ATTRIBUTES = [
        'sequence', 'name', 'hostname', 'role', 'memory', 'cpu', 'n_gpus', 'status'
    ]

    def __init__(self,
                 uuid,
                 sequence=None,
                 name=None,
                 hostname=None,
                 role=None,
                 docker_version=None,
                 kubelet_version=None,
                 os_image=None,
                 kernel_version=None,
                 schedulable_taints=None,
                 schedulable_state=None,
                 memory=None,
                 cpu=None,
                 n_gpus=None,
                 status=None,
                 gpus=None):
        self.uuid = uuid
        self.sequence = sequence
        self.name = name
        self.hostname = hostname
        self.role = role
        self.docker_version = docker_version
        self.kubelet_version = kubelet_version
        self.os_image = os_image
        self.kernel_version = kernel_version
        self.schedulable_taints = schedulable_taints
        self.schedulable_state = schedulable_state
        self.memory = memory
        self.cpu = cpu
        self.n_gpus = n_gpus
        self.status = status
        self.gpus = gpus


class PolyaxonClusterSchema(BaseSchema):
    version_api = fields.Dict()
    nodes = fields.Nested(ClusterNodeSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return PolyaxonClusterConfig


class PolyaxonClusterConfig(BaseConfig):
    """
    Polyaxon cluster config.

    Args:
        version_api: `dict`. The cluster's version api.
        nodes: list(ClusterNodeConfig). The nodes in the cluster.
    """
    SCHEMA = PolyaxonClusterSchema
    IDENTIFIER = 'PolyaxonCluster'

    def __init__(self, version_api, nodes=None):
        self.version_api = version_api
        self.nodes = nodes



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import copy
import numpy as np
import six

from marshmallow import fields, validates_schema
from marshmallow.exceptions import ValidationError

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.utils import (
    GeomSpace,
    LinSpace,
    LogNormal,
    LogSpace,
    LogUniform,
    Normal,
    PValue,
    QLogNormal,
    QLogUniform,
    QNormal,
    QUniform,
    Range,
    Uniform,
    lognormal,
    loguniform,
    normal,
    pvalues,
    qlognormal,
    qloguniform,
    qnormal,
    quniform,
    uniform,
    validate_pvalues
)

# pylint:disable=redefined-outer-name


def validate_matrix(values):
    v = sum(map(lambda x: 1 if x else 0, values))
    if v == 0 or v > 1:
        raise ValidationError("Matrix element is not valid, one and only one option is required.")


class MatrixSchema(BaseSchema):
    # Discrete
    values = fields.List(fields.Raw(), allow_none=True)
    pvalues = fields.List(PValue(), allow_none=True)
    range = Range(allow_none=True)
    linspace = LinSpace(allow_none=True)
    logspace = LogSpace(allow_none=True)
    geomspace = GeomSpace(allow_none=True)
    # Continuous
    uniform = Uniform(allow_none=True)
    quniform = QUniform(allow_none=True)
    loguniform = LogUniform(allow_none=True)
    qloguniform = QLogUniform(allow_none=True)
    normal = Normal(allow_none=True)
    qnormal = QNormal(allow_none=True)
    lognormal = LogNormal(allow_none=True)
    qlognormal = QLogNormal(allow_none=True)

    @staticmethod
    def schema_config():
        return MatrixConfig

    @validates_schema
    def validate_pvalues(self, data):
        if data.get('pvalues'):
            validate_pvalues(values=[v[1] for v in data['pvalues'] if v])

    @validates_schema
    def validate_matrix(self, data):
        validate_matrix([
            data.get('values'),
            data.get('pvalues'),
            data.get('range'),
            data.get('linspace'),
            data.get('logspace'),
            data.get('geomspace'),
            data.get('uniform'),
            data.get('quniform'),
            data.get('loguniform'),
            data.get('qloguniform'),
            data.get('normal'),
            data.get('qnormal'),
            data.get('lognormal'),
            data.get('qlognormal'),
        ])


class MatrixConfig(BaseConfig):
    IDENTIFIER = 'matrix'
    SCHEMA = MatrixSchema
    REDUCED_ATTRIBUTES = [
        'values', 'pvalues', 'range', 'linspace', 'logspace', 'geomspace',
        'uniform', 'quniform', 'loguniform', 'qloguniform',
        'normal', 'qnormal', 'lognormal', 'qlognormal'
    ]

    NUMPY_MAPPING = {
        'range': np.arange,
        'linspace': np.linspace,
        'logspace': np.logspace,
        'geomspace': np.geomspace,
        'uniform': uniform,
        'quniform': quniform,
        'loguniform': loguniform,
        'qloguniform': qloguniform,
        'normal': normal,
        'qnormal': qnormal,
        'lognormal': lognormal,
        'qlognormal': qlognormal,
    }

    RANGES = {
        'range', 'linspace', 'logspace', 'geomspace'
    }

    CONTINUOUS = {
        'uniform', 'quniform', 'loguniform', 'qloguniform',
        'normal', 'qnormal', 'lognormal', 'qlognormal'
    }

    DISTRIBUTIONS = {
        'pvalues',
        'uniform', 'quniform', 'loguniform', 'qloguniform',
        'normal', 'qnormal', 'lognormal', 'qlognormal'
    }

    def __init__(self,
                 values=None,
                 pvalues=None,
                 range=None,  # noqa
                 linspace=None,
                 logspace=None,
                 geomspace=None,
                 uniform=None,
                 quniform=None,
                 loguniform=None,
                 qloguniform=None,
                 normal=None,
                 qnormal=None,
                 lognormal=None,
                 qlognormal=None):
        self.values = values
        self.pvalues = pvalues
        self.range = range
        self.linspace = linspace
        self.logspace = logspace
        self.geomspace = geomspace
        self.uniform = uniform
        self.quniform = quniform
        self.loguniform = loguniform
        self.qloguniform = qloguniform
        self.normal = normal
        self.qnormal = qnormal
        self.lognormal = lognormal
        self.qlognormal = qlognormal

        validate_matrix([
            values, pvalues, range, linspace, logspace, geomspace, uniform, quniform,
            loguniform, qloguniform, normal, qnormal, lognormal, qlognormal])

    @property
    def is_distribution(self):
        key = list(six.iterkeys(self.to_dict()))[0]
        return key in self.DISTRIBUTIONS

    @property
    def is_continuous(self):
        key = list(six.iterkeys(self.to_dict()))[0]
        return key in self.CONTINUOUS

    @property
    def is_discrete(self):
        return not self.is_continuous

    @property
    def is_range(self):
        key = list(six.iterkeys(self.to_dict()))[0]
        return key in self.RANGES

    @property
    def is_categorical(self):
        key, value = list(six.iteritems(self.to_dict()))[0]
        if key != 'values':
            return False

        return any([v for v in value
                    if not isinstance(v, (int, float, complex, np.integer, np.floating))])

    @property
    def is_uniform(self):
        key = list(six.iterkeys(self.to_dict()))[0]
        return key == 'uniform'

    @property
    def min(self):
        if self.is_categorical:
            return None

        if self.is_range:
            value = list(six.itervalues(self.to_dict()))[0]
            return value.get('start')

        if self.is_discrete and not self.is_distribution:
            return min(self.to_numpy())

        if self.is_uniform:
            value = list(six.itervalues(self.to_dict()))[0]
            return value.get('low')

        return None

    @property
    def max(self):
        if self.is_categorical:
            return None

        if self.is_range:
            value = list(six.itervalues(self.to_dict()))[0]
            return value.get('stop')

        if self.is_discrete and not self.is_distribution:
            return max(self.to_numpy())

        if self.is_uniform:
            value = list(six.itervalues(self.to_dict()))[0]
            return value.get('high')

        return None

    @property
    def length(self):
        key, value = list(six.iteritems(self.to_dict()))[0]
        if key in ['values', 'pvalues']:
            return len(value)

        if key in self.DISTRIBUTIONS:
            raise ValidationError('Distribution should not call `to_numpy`, '
                                  'instead it should call `sample`.')

        return len(self.NUMPY_MAPPING[key](**value))

    def to_numpy(self):
        key, value = list(six.iteritems(self.to_dict()))[0]
        if key == 'values':
            return value

        if key in self.DISTRIBUTIONS:
            raise ValidationError('Distribution should not call `to_numpy`, '
                                  'instead it should call `sample`.')

        return self.NUMPY_MAPPING[key](**value)

    def sample(self, size=None, rand_generator=None):
        size = None if size == 1 else size
        key, value = list(six.iteritems(self.to_dict()))[0]
        value = copy.deepcopy(value)
        if key in {'values', 'range', 'linspace', 'logspace', 'geomspace'}:
            value = self.to_numpy()
            rand_generator = rand_generator or np.random
            try:
                return rand_generator.choice(value, size=size)
            except ValueError:
                idx = rand_generator.randint(0, len(value))
                return value[idx]

        if key == 'pvalues':
            return pvalues(values=value, size=size, rand_generator=rand_generator)

        value['size'] = size
        value['rand_generator'] = rand_generator
        return self.NUMPY_MAPPING[key](**value)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_schemas.base import BaseMultiSchema
from polyaxon_schemas.ml.layers.advanced_activations import (
    ELUConfig,
    LeakyReLUConfig,
    PReLUConfig,
    ThresholdedReLUConfig
)
from polyaxon_schemas.ml.layers.convolutional import (
    Conv1DConfig,
    Conv2DConfig,
    Conv2DTransposeConfig,
    Conv3DConfig,
    Conv3DTransposeConfig,
    Cropping1DConfig,
    Cropping2DConfig,
    Cropping3DConfig,
    SeparableConv2DConfig,
    UpSampling1DConfig,
    UpSampling2DConfig,
    UpSampling3DConfig,
    ZeroPadding1DConfig,
    ZeroPadding2DConfig,
    ZeroPadding3DConfig
)
from polyaxon_schemas.ml.layers.convolutional_recurrent import (
    ConvLSTM2DConfig,
    ConvRecurrent2DConfig
)
from polyaxon_schemas.ml.layers.core import (
    ActivationConfig,
    ActivityRegularizationConfig,
    CastConfig,
    DenseConfig,
    DropoutConfig,
    FlattenConfig,
    MaskingConfig,
    PermuteConfig,
    RepeatVectorConfig,
    ReshapeConfig,
    SpatialDropout1DConfig,
    SpatialDropout2DConfig,
    SpatialDropout3DConfig
)
from polyaxon_schemas.ml.layers.embeddings import EmbeddingConfig
from polyaxon_schemas.ml.layers.local import LocallyConnected1DConfig, LocallyConnected2DConfig
from polyaxon_schemas.ml.layers.merge import MergeConfig
from polyaxon_schemas.ml.layers.noise import (
    AlphaDropoutConfig,
    GaussianDropoutConfig,
    GaussianNoiseConfig
)
from polyaxon_schemas.ml.layers.normalization import BatchNormalizationConfig
from polyaxon_schemas.ml.layers.pooling import (
    AveragePooling1DConfig,
    AveragePooling2DConfig,
    AveragePooling3DConfig,
    GlobalAveragePooling1DConfig,
    GlobalAveragePooling2DConfig,
    GlobalAveragePooling3DConfig,
    GlobalMaxPooling1DConfig,
    GlobalMaxPooling2DConfig,
    GlobalMaxPooling3DConfig,
    MaxPooling1DConfig,
    MaxPooling2DConfig,
    MaxPooling3DConfig
)
from polyaxon_schemas.ml.layers.recurrent import (
    GRUConfig,
    LSTMConfig,
    RecurrentConfig,
    SimpleRNNConfig
)
from polyaxon_schemas.ml.layers.wrappers import (
    BidirectionalConfig,
    TimeDistributedConfig,
    WrapperConfig
)
from polyaxon_schemas.ml.processing.image import (
    AdjustBrightnessConfig,
    AdjustContrastConfig,
    AdjustGammaConfig,
    AdjustHueConfig,
    AdjustSaturationConfig,
    CentralCropConfig,
    ConvertColorSpaceConfig,
    ConvertImagesDtypeConfig,
    DrawBoundingBoxesConfig,
    ExtractGlimpseConfig,
    FlipConfig,
    RandomCropConfig,
    ResizeConfig,
    Rotate90Config,
    StandardizationConfig,
    ToBoundingBoxConfig,
    TotalVariationConfig,
    TransposeConfig
)


class LayerSchema(BaseMultiSchema):
    __multi_schema_name__ = 'layer'
    __configs__ = {
        LeakyReLUConfig.IDENTIFIER: LeakyReLUConfig,
        PReLUConfig.IDENTIFIER: PReLUConfig,
        ELUConfig.IDENTIFIER: ELUConfig,
        ThresholdedReLUConfig.IDENTIFIER: ThresholdedReLUConfig,
        Conv1DConfig.IDENTIFIER: Conv1DConfig,
        Conv2DConfig.IDENTIFIER: Conv2DConfig,
        Conv3DConfig.IDENTIFIER: Conv3DConfig,
        Conv2DTransposeConfig.IDENTIFIER: Conv2DTransposeConfig,
        Conv3DTransposeConfig.IDENTIFIER: Conv3DTransposeConfig,
        SeparableConv2DConfig.IDENTIFIER: SeparableConv2DConfig,
        UpSampling1DConfig.IDENTIFIER: UpSampling1DConfig,
        UpSampling2DConfig.IDENTIFIER: UpSampling2DConfig,
        UpSampling3DConfig.IDENTIFIER: UpSampling3DConfig,
        ZeroPadding1DConfig.IDENTIFIER: ZeroPadding1DConfig,
        ZeroPadding2DConfig.IDENTIFIER: ZeroPadding2DConfig,
        ZeroPadding3DConfig.IDENTIFIER: ZeroPadding3DConfig,
        Cropping1DConfig.IDENTIFIER: Cropping1DConfig,
        Cropping2DConfig.IDENTIFIER: Cropping2DConfig,
        Cropping3DConfig.IDENTIFIER: Cropping3DConfig,
        ConvRecurrent2DConfig.IDENTIFIER: ConvRecurrent2DConfig,
        ConvLSTM2DConfig.IDENTIFIER: ConvLSTM2DConfig,
        MaskingConfig.IDENTIFIER: MaskingConfig,
        DropoutConfig.IDENTIFIER: DropoutConfig,
        SpatialDropout1DConfig.IDENTIFIER: SpatialDropout1DConfig,
        SpatialDropout2DConfig.IDENTIFIER: SpatialDropout2DConfig,
        SpatialDropout3DConfig.IDENTIFIER: SpatialDropout3DConfig,
        ActivationConfig.IDENTIFIER: ActivationConfig,
        ReshapeConfig.IDENTIFIER: ReshapeConfig,
        PermuteConfig.IDENTIFIER: PermuteConfig,
        FlattenConfig.IDENTIFIER: FlattenConfig,
        RepeatVectorConfig.IDENTIFIER: RepeatVectorConfig,
        DenseConfig.IDENTIFIER: DenseConfig,
        ActivityRegularizationConfig.IDENTIFIER: ActivityRegularizationConfig,
        CastConfig.IDENTIFIER: CastConfig,
        EmbeddingConfig.IDENTIFIER: EmbeddingConfig,
        LocallyConnected1DConfig.IDENTIFIER: LocallyConnected1DConfig,
        LocallyConnected2DConfig.IDENTIFIER: LocallyConnected2DConfig,
        MergeConfig.IDENTIFIER: MergeConfig,
        GaussianNoiseConfig.IDENTIFIER: GaussianNoiseConfig,
        GaussianDropoutConfig.IDENTIFIER: GaussianDropoutConfig,
        AlphaDropoutConfig.IDENTIFIER: AlphaDropoutConfig,
        BatchNormalizationConfig.IDENTIFIER: BatchNormalizationConfig,
        MaxPooling1DConfig.IDENTIFIER: MaxPooling1DConfig,
        AveragePooling1DConfig.IDENTIFIER: AveragePooling1DConfig,
        MaxPooling2DConfig.IDENTIFIER: MaxPooling2DConfig,
        AveragePooling2DConfig.IDENTIFIER: AveragePooling2DConfig,
        MaxPooling3DConfig.IDENTIFIER: MaxPooling3DConfig,
        AveragePooling3DConfig.IDENTIFIER: AveragePooling3DConfig,
        GlobalAveragePooling1DConfig.IDENTIFIER: GlobalAveragePooling1DConfig,
        GlobalMaxPooling1DConfig.IDENTIFIER: GlobalMaxPooling1DConfig,
        GlobalAveragePooling2DConfig.IDENTIFIER: GlobalAveragePooling2DConfig,
        GlobalMaxPooling2DConfig.IDENTIFIER: GlobalMaxPooling2DConfig,
        GlobalAveragePooling3DConfig.IDENTIFIER: GlobalAveragePooling3DConfig,
        GlobalMaxPooling3DConfig.IDENTIFIER: GlobalMaxPooling3DConfig,
        RecurrentConfig.IDENTIFIER: RecurrentConfig,
        SimpleRNNConfig.IDENTIFIER: SimpleRNNConfig,
        GRUConfig.IDENTIFIER: GRUConfig,
        LSTMConfig.IDENTIFIER: LSTMConfig,
        WrapperConfig.IDENTIFIER: WrapperConfig,
        TimeDistributedConfig.IDENTIFIER: TimeDistributedConfig,
        BidirectionalConfig.IDENTIFIER: BidirectionalConfig,

        # Processors
        ResizeConfig.IDENTIFIER: ResizeConfig,
        CentralCropConfig.IDENTIFIER: CentralCropConfig,
        RandomCropConfig.IDENTIFIER: RandomCropConfig,
        ExtractGlimpseConfig.IDENTIFIER: ExtractGlimpseConfig,
        ToBoundingBoxConfig.IDENTIFIER: ToBoundingBoxConfig,
        FlipConfig.IDENTIFIER: FlipConfig,
        TransposeConfig.IDENTIFIER: TransposeConfig,
        Rotate90Config.IDENTIFIER: Rotate90Config,
        ConvertColorSpaceConfig.IDENTIFIER: ConvertColorSpaceConfig,
        ConvertImagesDtypeConfig.IDENTIFIER: ConvertImagesDtypeConfig,
        AdjustBrightnessConfig.IDENTIFIER: AdjustBrightnessConfig,
        AdjustContrastConfig.IDENTIFIER: AdjustContrastConfig,
        AdjustHueConfig.IDENTIFIER: AdjustHueConfig,
        AdjustSaturationConfig.IDENTIFIER: AdjustSaturationConfig,
        AdjustGammaConfig.IDENTIFIER: AdjustGammaConfig,
        StandardizationConfig.IDENTIFIER: StandardizationConfig,
        DrawBoundingBoxesConfig.IDENTIFIER: DrawBoundingBoxesConfig,
        TotalVariationConfig.IDENTIFIER: TotalVariationConfig,
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_schemas.specs.build import BuildSpecification
from polyaxon_schemas.specs.experiment import ExperimentSpecification
from polyaxon_schemas.specs.group import GroupSpecification
from polyaxon_schemas.specs.job import JobSpecification
from polyaxon_schemas.specs.notebook import NotebookSpecification
from polyaxon_schemas.specs.tensorboard import TensorboardSpecification

SPECIFICATION_BY_KIND = {
    BuildSpecification._SPEC_KIND: BuildSpecification,
    ExperimentSpecification._SPEC_KIND: ExperimentSpecification,
    GroupSpecification._SPEC_KIND: GroupSpecification,
    JobSpecification._SPEC_KIND: JobSpecification,
    NotebookSpecification._SPEC_KIND: NotebookSpecification,
    TensorboardSpecification._SPEC_KIND: TensorboardSpecification,
}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema, BaseSchema
from polyaxon_schemas.utils import ObjectOrListObject


class BaseBridgeSchema(BaseSchema):
    state_size = ObjectOrListObject(fields.Int, allow_none=True)
    name = fields.Str(allow_none=True)


class BaseBridgeConfig(BaseConfig):
    REDUCED_ATTRIBUTES = ['name']

    def __init__(self, state_size=None, name=None):
        self.state_size = state_size
        self.name = name


class LatentBridgeSchema(BaseBridgeSchema):
    latent_dim = fields.Int(allow_none=True)
    mean = fields.Float(allow_none=True)
    stddev = fields.Float(allow_none=True)

    @staticmethod
    def schema_config():
        return LatentBridgeConfig


class LatentBridgeConfig(BaseBridgeConfig):
    """A bridge that create a latent space based on the encoder output.

    This bridge could be used by VAE.

    Args(programmatic):
        mode: `str`, Specifies if this training, evaluation or prediction. See `Modes`.

    Args:
        latent_dim: `int`. The latent dimension to use.
        name: `str`. The name of this subgraph, used for creating the scope.

    Attributes:
        z_mean: `Tensor`. The latent distribution mean.
        z_log_sigma: `Tensor`. The latent distribution log variance.

    Returns:
        `BridgeSpec`

    Programmatic usage:

    ```python
    def bridge_fn(mode, features, labels, loss, encoder_fn, decoder_fn):
        return plx.bridges.LatentBridge(mode)(features, labels, loss, encoder_fn, decoder_fn)
    ```

    Polyaxonfile usage:

    ```yaml
    model:
      generator:
        ...
        bridge: LatentBridge
        encoder:
          input_layers: image
          layers:
            - Dense:
                units: 128
            - Dense:
                units: 256
                name: encoded
        decoder:
          input_layers: encoded
          layers:
            - Dense:
                units: 256
            - Dense:
                units: 784
    ```
    """
    IDENTIFIER = 'LatentBridge'
    SCHEMA = LatentBridgeSchema

    def __init__(self, latent_dim=1, mean=0., stddev=1., **kwargs):
        super(LatentBridgeConfig, self).__init__(**kwargs)
        self.latent_dim = latent_dim
        self.mean = mean
        self.stddev = stddev


class NoOpBridgeSchema(BaseBridgeSchema):

    @staticmethod
    def schema_config():
        return NoOpBridgeConfig


class NoOpBridgeConfig(BaseBridgeConfig):
    """A bridge that passes the encoder output to the decoder.

    This bridge could be used by VAE.

    Args(programmatic):
        mode: `str`, Specifies if this training, evaluation or prediction. See `Modes`.

    Args:
        state_size: `int`. The latent dimension to use.
        name: `str`. The name of this subgraph, used for creating the scope.

    Returns:
        `BridgeSpec`

    Programmatic usage:

    ```python
    def bridge_fn(mode, features, labels, loss, encoder_fn, decoder_fn):
        return plx.bridges.NoOpBridge(mode)(features, labels, loss, encoder_fn, decoder_fn)
    ```

    Polyaxonfile usage:

    ```yaml
    model:
      generator:
        ...
        bridge: NoOpBridge
        encoder:
          input_layers: image
          layers:
            - Dense:
                units: 128
            - Dense:
                units: 256
                name: encoded
        decoder:
          input_layers: encoded
          layers:
            - Dense:
                units: 256
            - Dense:
                units: 784
    ```
    """
    IDENTIFIER = 'NoOpBridge'
    SCHEMA = NoOpBridgeSchema


class BridgeSchema(BaseMultiSchema):
    __multi_schema_name__ = 'bridge'
    __configs__ = {
        LatentBridgeConfig.IDENTIFIER: LatentBridgeConfig,
        NoOpBridgeConfig.IDENTIFIER: NoOpBridgeConfig,
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.utils import ObjectOrListObject


class RunSchema(BaseSchema):
    cmd = ObjectOrListObject(fields.Str)

    @staticmethod
    def schema_config():
        return RunConfig


class RunConfig(BaseConfig):
    SCHEMA = RunSchema
    IDENTIFIER = 'run'

    def __init__(self, cmd):
        self.cmd = cmd



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.ml.hooks import HookSchema
from polyaxon_schemas.ml.processing.pipelines import PipelineSchema


class EvalSchema(BaseSchema):
    data_pipeline = fields.Nested(PipelineSchema)
    steps = fields.Int(allow_none=True)
    hooks = fields.Nested(HookSchema, many=True, allow_none=True)
    delay_secs = fields.Int(allow_none=True)
    continuous_eval_throttle_secs = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return EvalConfig


class EvalConfig(BaseConfig):
    SCHEMA = EvalSchema
    IDENTIFIER = 'eval'

    def __init__(self,
                 data_pipeline,
                 steps=10,
                 hooks=None,
                 delay_secs=0,
                 continuous_eval_throttle_secs=60):
        self.data_pipeline = data_pipeline
        self.steps = steps
        self.hooks = hooks
        self.delay_secs = delay_secs
        self.continuous_eval_throttle_secs = continuous_eval_throttle_secs



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import tempfile

from unittest import TestCase

from polyaxon_schemas.polyaxonfile import (
    DEFAULT_POLYAXON_FILE_EXTENSION,
    DEFAULT_POLYAXON_FILE_NAME,
    PolyaxonFile
)


class TestDefaultFile(TestCase):
    def test_default_not_found(self):
        path = tempfile.mkdtemp()
        assert PolyaxonFile.check_default_path(path=path) is None

    def test_polyaxon_found(self):
        def create_file(path, filename, ext):
            fpath = '{}/{}.{}'.format(path, filename, ext)
            open(fpath, 'w')

        for filename in DEFAULT_POLYAXON_FILE_NAME:
            for ext in DEFAULT_POLYAXON_FILE_EXTENSION:
                path = tempfile.mkdtemp()
                create_file(path, filename, ext)
                assert PolyaxonFile.check_default_path(path=path)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema, BaseSchema
from polyaxon_schemas.utils import Tensor

# pylint:disable=too-many-lines


class BaseMetricSchema(BaseSchema):
    input_layer = Tensor(allow_none=True)
    output_layer = Tensor(allow_none=True)
    weights = fields.Float(allow_none=True)
    metrics_collections = fields.List(fields.Str(), allow_none=True)
    updates_collections = fields.List(fields.Str(), allow_none=True)
    name = fields.Str(allow_none=True)

    @staticmethod
    def schema_config():
        return BaseMetricConfig


class BaseMetricConfig(BaseConfig):
    REDUCED_ATTRIBUTES = ['input_layer', 'output_layer', 'weights',
                          'metrics_collections', 'updates_collections', 'name']

    def __init__(self,
                 input_layer=None,
                 output_layer=None,
                 weights=None,
                 metrics_collections=None,
                 updates_collections=None,
                 name=None):
        self.input_layer = input_layer
        self.output_layer = output_layer
        self.weights = weights
        self.metrics_collections = metrics_collections
        self.updates_collections = updates_collections
        self.name = name


class BaseTensorMetricSchema(BaseSchema):
    values = Tensor()
    weights = fields.Float(allow_none=True)
    metrics_collections = fields.List(fields.Str(), allow_none=True)
    updates_collections = fields.List(fields.Str(), allow_none=True)
    name = fields.Str(allow_none=True)

    @staticmethod
    def schema_config():
        return BaseTensorMetricConfig


class BaseTensorMetricConfig(BaseConfig):
    REDUCED_ATTRIBUTES = ['weights', 'metrics_collections', 'updates_collections', 'name']

    def __init__(self,
                 values,
                 weights=None,
                 metrics_collections=None,
                 updates_collections=None,
                 name=None):
        self.values = values
        self.weights = weights
        self.metrics_collections = metrics_collections
        self.updates_collections = updates_collections
        self.name = name


class TruePositivesSchema(BaseMetricSchema):

    @staticmethod
    def schema_config():
        return TruePositivesConfig


class TruePositivesConfig(BaseMetricConfig):
    """Sum the weights of true_positives.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: The input true values layer, defaults to labels. Will
          be cast to `bool`.
        output_layer: The output layer generated by the network,
            default to last layer of the network. Will be cast to `bool`.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions
            must be either `1`, or the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that `mean`
            should be added to.
        updates_collections: An optional list of collections that `update_op`
            should be added to.
        name: An optional variable_scope name.

    Returns:
        value_tensor: A `Tensor` representing the current value of the metric.
        update_op: An operation that accumulates the error from a batch of data.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - TruePositives
      # other model properties
    ```
    """
    IDENTIFIER = 'TruePositives'
    SCHEMA = TruePositivesSchema


class TrueNegativesSchema(BaseMetricSchema):

    @staticmethod
    def schema_config():
        return TrueNegativesConfig


class TrueNegativesConfig(BaseMetricConfig):
    """Sum the weights of true_negatives.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: The input true values layer, defaults to labels. Will
          be cast to `bool`.
        output_layer: The output layer generated by the network,
            default to last layer of the network. Will be cast to `bool`.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions
            must be either `1`, or the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that `mean`
            should be added to.
        updates_collections: An optional list of collections that `update_op`
            should be added to.
        name: An optional variable_scope name.

    Returns:
        value_tensor: A `Tensor` representing the current value of the metric.
        update_op: An operation that accumulates the error from a batch of data.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
           `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - TruePositives
        - TrueNegatives:
            input_layer: labels_1
            output_layer: dense_out_3
      # other model properties
    ```
    """
    IDENTIFIER = 'TrueNegatives'
    SCHEMA = TrueNegativesSchema


class FalsePositivesSchema(BaseMetricSchema):

    @staticmethod
    def schema_config():
        return FalsePositivesConfig


class FalsePositivesConfig(BaseMetricConfig):
    """Sum the weights of false positives.
    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: The input true values layer, defaults to labels. Will
          be cast to `bool`.
        output_layer: The output layer generated by the network,
            default to last layer of the network. Will be cast to `bool`.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions
            must be either `1`, or the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that `mean`
            should be added to.
        updates_collections: An optional list of collections that `update_op`
            should be added to.
        name: An optional variable_scope name.

    Returns:
        value_tensor: A `Tensor` representing the current value of the metric.
        update_op: An operation that accumulates the error from a batch of data.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - FalsePositives
      # other model properties
    ```
    """
    IDENTIFIER = 'FalsePositives'
    SCHEMA = FalsePositivesSchema


class FalseNegativesSchema(BaseMetricSchema):

    @staticmethod
    def schema_config():
        return FalseNegativesConfig


class FalseNegativesConfig(BaseMetricConfig):
    """Computes the total number of false negatives.
    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: The input true values layer, defaults to labels. Will
          be cast to `bool`.
        output_layer: The output layer generated by the network,
            default to last layer of the network. Will be cast to `bool`.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions
            must be either `1`, or the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that `mean`
            should be added to.
        updates_collections: An optional list of collections that `update_op`
            should be added to.
        name: An optional variable_scope name.

    Returns:
        value_tensor: A `Tensor` representing the current value of the metric.
        update_op: An operation that accumulates the error from a batch of data.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - TruePositives
        - TrueNegatives
        - FalseNegatives:
            input_layer: labels_1
            output_layer: dense_out_3
      # other model properties
    ```
    """
    IDENTIFIER = 'FalseNegatives'
    SCHEMA = FalseNegativesSchema


class MeanSchema(BaseTensorMetricSchema):

    @staticmethod
    def schema_config():
        return MeanConfig


class MeanConfig(BaseTensorMetricConfig):
    """Computes the (weighted) mean of the given values.

    The `streaming_mean` function creates two local variables, `total` and `count`
    that are used to compute the average of `values`. This average is ultimately
    returned as `mean` which is an idempotent operation that simply divides
    `total` by `count`.

    For estimation of the metric  over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the `mean`.
    `update_op` increments `total` with the reduced sum of the product of `values`
    and `weights`, and it increments `count` with the reduced sum of `weights`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        values: A `Tensor` of arbitrary dimensions.
        weights: `Tensor` whose rank is either 0, or the same rank as `values`, and
            must be broadcastable to `values` (i.e., all dimensions must be either
            `1`, or the same as the corresponding `values` dimension).
        metrics_collections: An optional list of collections that `mean`
            should be added to.
        updates_collections: An optional list of collections that `update_op`
            should be added to.
        name: An optional variable_scope name.

    Returns:
        mean: A `Tensor` representing the current mean, the value of `total` divided
          by `count`.
        update_op: An operation that increments the `total` and `count` variables
          appropriately and whose value matches `mean`.

    Raises:
        ValueError: If `weights` is not `None` and its shape doesn't match `values`,
            or if either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - Mean:
            values: layer1
      # other model properties
    ```
    """
    IDENTIFIER = 'Mean'
    SCHEMA = MeanSchema


class MeanTensorSchema(BaseTensorMetricSchema):

    @staticmethod
    def schema_config():
        return MeanTensorConfig


class MeanTensorConfig(BaseTensorMetricConfig):
    """Computes the element-wise (weighted) mean of the given tensors.

    In contrast to the `streaming_mean` function which returns a scalar with the
    mean,  this function returns an average tensor with the same shape as the
    input tensors.

    The `streaming_mean_tensor` function creates two local variables,
    `total_tensor` and `count_tensor` that are used to compute the average of
    `values`. This average is ultimately returned as `mean` which is an idempotent
    operation that simply divides `total` by `count`.

    For estimation of the metric  over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the `mean`.
    `update_op` increments `total` with the reduced sum of the product of `values`
    and `weights`, and it increments `count` with the reduced sum of `weights`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        values: A `Tensor` of arbitrary dimensions.
        weights: `Tensor` whose rank is either 0, or the same rank as `values`, and
            must be broadcastable to `values` (i.e., all dimensions must be either
            `1`, or the same as the corresponding `values` dimension).
        metrics_collections: An optional list of collections that `mean`
            should be added to.
        updates_collections: An optional list of collections that `update_op`
            should be added to.
        name: An optional variable_scope name.

    Returns:
      mean: A float `Tensor` representing the current mean, the value of `total`
        divided by `count`.
      update_op: An operation that increments the `total` and `count` variables
        appropriately and whose value matches `mean`.

    Raises:
        ValueError: If `weights` is not `None` and its shape doesn't match `values`,
            or if either `metrics_collections` or `updates_collections` are not a list or tuple.

     Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - MeanTensor:
            values: layer1
      # other model properties
    ```
    """
    IDENTIFIER = 'MeanTensor'
    SCHEMA = MeanTensorSchema


class AccuracySchema(BaseMetricSchema):

    @staticmethod
    def schema_config():
        return AccuracyConfig


class AccuracyConfig(BaseMetricConfig):
    """Calculates how often `predictions` matches `labels`.

    The `streaming_accuracy` function creates two local variables, `total` and
    `count` that are used to compute the frequency with which `predictions`
    matches `labels`. This frequency is ultimately returned as `accuracy`: an
    idempotent operation that simply divides `total` by `count`.

    For estimation of the metric  over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the `accuracy`.
    Internally, an `is_correct` operation computes a `Tensor` with elements 1.0
    where the corresponding elements of `predictions` and `labels` match and 0.0
    otherwise. Then `update_op` increments `total` with the reduced sum of the
    product of `weights` and `is_correct`, and it increments `count` with the
    reduced sum of `weights`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: The input true values layer, defaults to labels. Will
          be cast to `bool`.
        output_layer: The output layer generated by the network,
            default to last layer of the network. Will be cast to `bool`.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions
            must be either `1`, or the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that `mean`
            should be added to.
        updates_collections: An optional list of collections that `update_op`
            should be added to.
        name: An optional variable_scope name.

    Returns:
        accuracy: A `Tensor` representing the accuracy, the value of `total` divided by `count`.
        update_op: An operation that increments the `total` and `count` variables
            appropriately and whose value matches `accuracy`.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - Precision
        - Accuracy:
            input_layer: labels_1
            output_layer: dense_out_3
      # other model properties
    ```
    """
    IDENTIFIER = 'Accuracy'
    SCHEMA = AccuracySchema


class PrecisionSchema(BaseMetricSchema):

    @staticmethod
    def schema_config():
        return PrecisionConfig


class PrecisionConfig(BaseMetricConfig):
    """Computes the precision of the predictions with respect to the labels.

    The `streaming_precision` function creates two local variables,
    `true_positives` and `false_positives`, that are used to compute the
    precision. This value is ultimately returned as `precision`, an idempotent
    operation that simply divides `true_positives` by the sum of `true_positives`
    and `false_positives`.

    For estimation of the metric  over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the
    `precision`. `update_op` weights each prediction by the corresponding value in
    `weights`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: The input true values layer, defaults to labels. Will
          be cast to `bool`.
        output_layer: The output layer generated by the network,
            default to last layer of the network. Will be cast to `bool`.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions
            must be either `1`, or the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that `mean`
            should be added to.
        updates_collections: An optional list of collections that `update_op`
            should be added to.
        name: An optional variable_scope name.

    Returns:
        precision: Scalar float `Tensor` with the value of `true_positives`
            divided by the sum of `true_positives` and `false_positives`.
        update_op: `Operation` that increments `true_positives` and
            `false_positives` variables appropriately and whose value matches `precision`.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - Precision
        - Accuracy
      # other model properties
    ```
    """
    IDENTIFIER = 'Precision'
    SCHEMA = PrecisionSchema


class RecallSchema(BaseMetricSchema):

    @staticmethod
    def schema_config():
        return RecallConfig


class RecallConfig(BaseMetricConfig):
    """Computes the recall of the predictions with respect to the labels.

    The `streaming_recall` function creates two local variables, `true_positives`
    and `false_negatives`, that are used to compute the recall. This value is
    ultimately returned as `recall`, an idempotent operation that simply divides
    `true_positives` by the sum of `true_positives`  and `false_negatives`.

    For estimation of the metric  over a stream of data, the function creates an
    `update_op` that updates these variables and returns the `recall`. `update_op`
    weights each prediction by the corresponding value in `weights`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: The input true values layer, defaults to labels. Will
          be cast to `bool`.
        output_layer: The output layer generated by the network,
            default to last layer of the network. Will be cast to `bool`.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions
            must be either `1`, or the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that `mean`
            should be added to.
        updates_collections: An optional list of collections that `update_op`
            should be added to.
        name: An optional variable_scope name.

    Returns:
        recall: Scalar float `Tensor` with the value of `true_positives` divided
            by the sum of `true_positives` and `false_negatives`.
        update_op: `Operation` that increments `true_positives` and
            `false_negatives` variables appropriately and whose value matches `recall`.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - Precision
        - Accuracy
        - Recall
      # other model properties
    ```
    """
    IDENTIFIER = 'Recall'
    SCHEMA = RecallSchema


class AUCSchema(BaseMetricSchema):
    num_thresholds = fields.Int(allow_none=True)
    curve = fields.Str(allow_none=True)

    @staticmethod
    def schema_config():
        return AUCConfig


class AUCConfig(BaseMetricConfig):
    """Computes the approximate AUC via a Riemann sum.

    The `streaming_auc` function creates four local variables, `true_positives`,
    `true_negatives`, `false_positives` and `false_negatives` that are used to
    compute the AUC. To discretize the AUC curve, a linearly spaced set of
    thresholds is used to compute pairs of recall and precision values. The area
    under the ROC-curve is therefore computed using the height of the recall
    values by the false positive rate, while the area under the PR-curve is the
    computed using the height of the precision values by the recall.

    This value is ultimately returned as `auc`, an idempotent operation that
    computes the area under a discretized curve of precision versus recall values
    (computed using the aforementioned variables). The `num_thresholds` variable
    controls the degree of discretization with larger numbers of thresholds more
    closely approximating the true AUC. The quality of the approximation may vary
    dramatically depending on `num_thresholds`.

    For best results, `predictions` should be distributed approximately uniformly
    in the range [0, 1] and not peaked around 0 or 1. The quality of the AUC
    approximation may be poor if this is not the case.

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the `auc`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: The input true values layer, defaults to labels. Will
          be cast to `bool`.
        output_layer: The output layer generated by the network,
            default to last layer of the network. Will be cast to `bool`.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions
            must be either `1`, or the same as the corresponding `labels` dimension).
        num_thresholds: The number of thresholds to use when discretizing the roc curve.
        metrics_collections: An optional list of collections that `mean`
            should be added to.
        updates_collections: An optional list of collections that `update_op`
            should be added to.
        curve: Specifies the name of the curve to be computed, 'ROC' [default] or
            'PR' for the Precision-Recall-curve.
        name: An optional variable_scope name.

    Returns:
      auc: A scalar `Tensor` representing the current area-under-curve.
      update_op: An operation that increments the `true_positives`,
        `true_negatives`, `false_positives` and `false_negatives` variables
        appropriately and whose value matches `auc`.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - AUC
      # other model properties
    ```
    """
    IDENTIFIER = 'AUC'
    SCHEMA = AUCSchema

    def __init__(self, num_thresholds=200, curve='ROC', **kwargs):
        super(AUCConfig, self).__init__(**kwargs)
        self.num_thresholds = num_thresholds
        self.curve = curve


class SpecificityAtSensitivitySchema(BaseMetricSchema):
    sensitivity = fields.Float(validate=validate.Range(min=0., max=1.))
    num_thresholds = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return SpecificityAtSensitivityConfig


class SpecificityAtSensitivityConfig(BaseMetricConfig):
    """Computes the specificity at a given sensitivity.

    The `streaming_specificity_at_sensitivity` function creates four local
    variables, `true_positives`, `true_negatives`, `false_positives` and
    `false_negatives` that are used to compute the specificity at the given
    sensitivity value. The threshold for the given sensitivity value is computed
    and used to evaluate the corresponding specificity.

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the
    `specificity`. `update_op` increments the `true_positives`, `true_negatives`,
    `false_positives` and `false_negatives` counts with the weight of each case
    found in the `predictions` and `labels`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    For additional information about specificity and sensitivity, see the
    following: https://en.wikipedia.org/wiki/Sensitivity_and_specificity

    Args:
        input_layer: The input true values layer, defaults to labels. Will
          be cast to `bool`.
        output_layer: The output layer generated by the network,
            default to last layer of the network. Will be cast to `bool`.
            If the network has multiple output, you should specify which layer to use.
        sensitivity: A scalar value in range `[0, 1]`.
        weights: `Tensor` whose rank is either 0, or the same rank as `labels`, and
            must be broadcastable to `labels` (i.e., all dimensions must be either
            `1`, or the same as the corresponding `labels` dimension).
        num_thresholds: The number of thresholds to use for matching the given sensitivity.
        metrics_collections: An optional list of collections that `specificity`
            should be added to.
        updates_collections: An optional list of collections that `update_op` should
            be added to.
        name: An optional variable_scope name.

    Returns:
        specificity: A scalar `Tensor` representing the specificity at the given
            `specificity` value.
        update_op: An operation that increments the `true_positives`,
            `true_negatives`, `false_positives` and `false_negatives` variables
            appropriately and whose value matches `specificity`.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            `sensitivity` is not between 0 and 1, or if either `metrics_collections`
            or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - SpecificityAtSensitivity
      # other model properties
    ```
    """
    IDENTIFIER = 'SpecificityAtSensitivity'
    SCHEMA = SpecificityAtSensitivitySchema

    def __init__(self, sensitivity, num_thresholds=200, **kwargs):
        self.sensitivity = sensitivity
        self.num_thresholds = num_thresholds
        super(SpecificityAtSensitivityConfig, self).__init__(**kwargs)


class SensitivityAtSpecificitySchema(BaseMetricSchema):
    specificity = fields.Float(validate=validate.Range(min=0., max=1.))
    num_thresholds = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return SensitivityAtSpecificityConfig


class SensitivityAtSpecificityConfig(BaseMetricConfig):
    """Computes the specificity at a given sensitivity.

    The `streaming_sensitivity_at_specificity` function creates four local
    variables, `true_positives`, `true_negatives`, `false_positives` and
    `false_negatives` that are used to compute the sensitivity at the given
    specificity value. The threshold for the given specificity value is computed
    and used to evaluate the corresponding sensitivity.

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the
    `sensitivity`. `update_op` increments the `true_positives`, `true_negatives`,
    `false_positives` and `false_negatives` counts with the weight of each case
    found in the `predictions` and `labels`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    For additional information about specificity and sensitivity, see the
    following: https://en.wikipedia.org/wiki/Sensitivity_and_specificity

    Args:
        input_layer: The input true values layer, defaults to labels. Will
          be cast to `bool`.
        output_layer: The output layer generated by the network,
            default to last layer of the network. Will be cast to `bool`.
            If the network has multiple output, you should specify which layer to use.
        specificity: A scalar value in range `[0, 1]`.
        weights: `Tensor` whose rank is either 0, or the same rank as `labels`, and
            must be broadcastable to `labels` (i.e., all dimensions must be either
            `1`, or the same as the corresponding `labels` dimension).
        num_thresholds: The number of thresholds to use for matching the given sensitivity.
        metrics_collections: An optional list of collections that `specificity`
            should be added to.
        updates_collections: An optional list of collections that `update_op` should
            be added to.
        name: An optional variable_scope name.

    Returns:
        sensitivity: A scalar `Tensor` representing the sensitivity at the given
            `specificity` value.
        update_op: An operation that increments the `true_positives`,
            `true_negatives`, `false_positives` and `false_negatives` variables
            appropriately and whose value matches `sensitivity`.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            `specificity` is not between 0 and 1, or if either `metrics_collections`
            or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - SpecificityAtSensitivity
      # other model properties
    ```
    """
    IDENTIFIER = 'SensitivityAtSpecificity'
    SCHEMA = SensitivityAtSpecificitySchema

    def __init__(self, specificity, num_thresholds=200, **kwargs):
        self.specificity = specificity
        self.num_thresholds = num_thresholds
        super(SensitivityAtSpecificityConfig, self).__init__(**kwargs)


class PrecisionAtThresholdsSchema(BaseMetricSchema):
    thresholds = fields.List(fields.Float(validate=validate.Range(min=0., max=1.)))

    @staticmethod
    def schema_config():
        return PrecisionAtThresholdsConfig


class PrecisionAtThresholdsConfig(BaseMetricConfig):
    """Computes precision values for different `thresholds` on `predictions`.

    The `streaming_precision_at_thresholds` function creates four local variables,
    `true_positives`, `true_negatives`, `false_positives` and `false_negatives`
    for various values of thresholds. `precision[i]` is defined as the total
    weight of values in `predictions` above `thresholds[i]` whose corresponding
    entry in `labels` is `True`, divided by the total weight of values in
    `predictions` above `thresholds[i]` (`true_positives[i] / (true_positives[i] +
    false_positives[i])`).

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the
    `precision`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: The input true values layer, defaults to labels. Will
          be cast to `bool`.
        output_layer: The output layer generated by the network,
            default to last layer of the network. Will be cast to `bool`.
            If the network has multiple output, you should specify which layer to use.
        thresholds: A python list or tuple of float thresholds in `[0, 1]`.
        weights: `Tensor` whose rank is either 0, or the same rank as `labels`, and
            must be broadcastable to `labels` (i.e., all dimensions must be either
            `1`, or the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that `specificity`
            should be added to.
        updates_collections: An optional list of collections that `update_op` should
            be added to.
        name: An optional variable_scope name.

    Returns:
        precision: A float `Tensor` of shape `[len(thresholds)]`.
        update_op: An operation that increments the `true_positives`,
            `true_negatives`, `false_positives` and `false_negatives` variables that
            are used in the computation of `precision`.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - PrecisionAtThresholds: {thresholds: [0.1, 0.1, 0.1]}
      # other model properties
    ```
    """
    IDENTIFIER = 'PrecisionAtThresholds'
    SCHEMA = PrecisionAtThresholdsSchema

    def __init__(self, thresholds, **kwargs):
        self.thresholds = thresholds
        super(PrecisionAtThresholdsConfig, self).__init__(**kwargs)


class RecallAtThresholdsSchema(BaseMetricSchema):
    thresholds = fields.List(fields.Float(validate=validate.Range(min=0., max=1.)))

    @staticmethod
    def schema_config():
        return RecallAtThresholdsConfig


class RecallAtThresholdsConfig(BaseMetricConfig):
    """Computes various recall values for different `thresholds` on `predictions`.

    The `streaming_recall_at_thresholds` function creates four local variables,
    `true_positives`, `true_negatives`, `false_positives` and `false_negatives`
    for various values of thresholds. `recall[i]` is defined as the total weight
    of values in `predictions` above `thresholds[i]` whose corresponding entry in
    `labels` is `True`, divided by the total weight of `True` values in `labels`
    (`true_positives[i] / (true_positives[i] + false_negatives[i])`).

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the `recall`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: The input true values layer, defaults to labels. Will
          be cast to `bool`.
        output_layer: The output layer generated by the network,
            default to last layer of the network. Will be cast to `bool`.
            If the network has multiple output, you should specify which layer to use.
        thresholds: A python list or tuple of float thresholds in `[0, 1]`.
        weights: `Tensor` whose rank is either 0, or the same rank as `labels`, and
            must be broadcastable to `labels` (i.e., all dimensions must be either
            `1`, or the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that `specificity`
            should be added to.
        updates_collections: An optional list of collections that `update_op` should
            be added to.
        name: An optional variable_scope name.

    Returns:
        precision: A float `Tensor` of shape `[len(thresholds)]`.
        update_op: An operation that increments the `true_positives`,
            `true_negatives`, `false_positives` and `false_negatives` variables that
            are used in the computation of `precision`.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - RecallAtThresholds: {thresholds: [0.1, 0.1, 0.1]}
      # other model properties
    ```
    """
    IDENTIFIER = 'RecallAtThresholds'
    SCHEMA = RecallAtThresholdsSchema

    def __init__(self, thresholds, **kwargs):
        self.thresholds = thresholds
        super(RecallAtThresholdsConfig, self).__init__(**kwargs)


class SparseRecallAtKSchema(BaseMetricSchema):
    k = fields.Int()
    class_id = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return SparseRecallAtKConfig


class SparseRecallAtKConfig(BaseMetricConfig):
    """Computes recall@k of the predictions with respect to sparse labels.

    If `class_id` is not specified, we'll calculate recall as the ratio of true
        positives (i.e., correct predictions, items in the top `k` highest
        `predictions` that are found in the corresponding row in `labels`) to
        actual positives (the full `labels` row).
    If `class_id` is specified, we calculate recall by considering only the rows
        in the batch for which `class_id` is in `labels`, and computing the
        fraction of them for which `class_id` is in the corresponding row in
        `labels`.

    `streaming_sparse_recall_at_k` creates two local variables,
    `true_positive_at_<k>` and `false_negative_at_<k>`, that are used to compute
    the recall_at_k frequency. This frequency is ultimately returned as
    `recall_at_<k>`: an idempotent operation that simply divides
    `true_positive_at_<k>` by total (`true_positive_at_<k>` +
    `false_negative_at_<k>`).

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the
    `recall_at_<k>`. Internally, a `top_k` operation computes a `Tensor`
    indicating the top `k` `predictions`. Set operations applied to `top_k` and
    `labels` calculate the true positives and false negatives weighted by
    `weights`. Then `update_op` increments `true_positive_at_<k>` and
    `false_negative_at_<k>` using these values.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: `int64` `Tensor` or `SparseTensor` with shape
            [D1, ... DN, num_labels], where N >= 1 and num_labels is the number of
            target classes for the associated prediction. Commonly, N=1 and `labels`
            has shape [batch_size, num_labels]. [D1, ... DN] must match `predictions`.
            Values should be in range [0, num_classes), where num_classes is the last
            dimension of `predictions`. Values outside this range always count
            towards `false_negative_at_<k>`.
        output_layer: Float `Tensor` with shape [D1, ... DN, num_classes] where
            N >= 1. Commonly, N=1 and predictions has shape [batch size, num_classes].
            The final dimension contains the logit values for each class. [D1, ... DN]
            must match `labels`.
        k: Integer, k for @k metric.
        class_id: Integer class ID for which we want binary metrics. This should be
            in range [0, num_classes), where num_classes is the last dimension of
            `predictions`. If class_id is outside this range, the method returns NAN.
        weights: `Tensor` whose rank is either 0, or n-1, where n is the rank of
            `labels`. If the latter, it must be broadcastable to `labels` (i.e., all
            dimensions must be either `1`, or the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that values should be added to.
        updates_collections: An optional list of collections that updates should be added to.
        name: Name of new update operation, and namespace for other dependent ops.

    Returns:
        recall: Scalar `float64` `Tensor` with the value of `true_positives` divided
            by the sum of `true_positives` and `false_negatives`.
        update_op: `Operation` that increments `true_positives` and
            `false_negatives` variables appropriately, and whose value matches `recall`.

    Raises:
        ValueError: If `weights` is not `None` and its shape doesn't match
        `predictions`, or if either `metrics_collections` or `updates_collections`
        are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - SparseRecallAtK: {k: 2}
      # other model properties
    ```
    """
    IDENTIFIER = 'SparseRecallAtK'
    SCHEMA = SparseRecallAtKSchema

    def __init__(self, k, class_id=None, **kwargs):
        self.k = k
        self.class_id = class_id
        super(SparseRecallAtKConfig, self).__init__(**kwargs)


class SparsePrecisionAtKSchema(BaseMetricSchema):
    k = fields.Int()
    class_id = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return SparsePrecisionAtKConfig


class SparsePrecisionAtKConfig(BaseMetricConfig):
    """Computes precision@k of the predictions with respect to sparse labels.

    If `class_id` is not specified, we calculate precision as the ratio of true
        positives (i.e., correct predictions, items in the top `k` highest
        `predictions` that are found in the corresponding row in `labels`) to
        positives (all top `k` `predictions`).
    If `class_id` is specified, we calculate precision by considering only the
        rows in the batch for which `class_id` is in the top `k` highest
        `predictions`, and computing the fraction of them for which `class_id` is
        in the corresponding row in `labels`.

    We expect precision to decrease as `k` increases.

    `streaming_sparse_precision_at_k` creates two local variables,
    `true_positive_at_<k>` and `false_positive_at_<k>`, that are used to compute
    the precision@k frequency. This frequency is ultimately returned as
    `precision_at_<k>`: an idempotent operation that simply divides
    `true_positive_at_<k>` by total (`true_positive_at_<k>` +
    `false_positive_at_<k>`).

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the
    `precision_at_<k>`. Internally, a `top_k` operation computes a `Tensor`
    indicating the top `k` `predictions`. Set operations applied to `top_k` and
    `labels` calculate the true positives and false positives weighted by
    `weights`. Then `update_op` increments `true_positive_at_<k>` and
    `false_positive_at_<k>` using these values.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: `int64` `Tensor` or `SparseTensor` with shape
            [D1, ... DN, num_labels], where N >= 1 and num_labels is the number of
            target classes for the associated prediction. Commonly, N=1 and `labels`
            has shape [batch_size, num_labels]. [D1, ... DN] must match
            `predictions`. Values should be in range [0, num_classes), where
            num_classes is the last dimension of `predictions`. Values outside this
            range are ignored.
        output_layer: Float `Tensor` with shape [D1, ... DN, num_classes] where
            N >= 1. Commonly, N=1 and predictions has shape [batch size, num_classes].
            The final dimension contains the logit values for each class. [D1, ... DN]
            must match `labels`.
        k: Integer, k for @k metric.
        class_id: Integer class ID for which we want binary metrics. This should be
            in range [0, num_classes], where num_classes is the last dimension of
            `predictions`. If `class_id` is outside this range, the method returns NAN.
        weights: `Tensor` whose rank is either 0, or n-1, where n is the rank of
            `labels`. If the latter, it must be broadcastable to `labels` (i.e., all
            dimensions must be either `1`, or the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that values should
            be added to.
        updates_collections: An optional list of collections that updates should be added to.
        name: Name of new update operation, and namespace for other dependent ops.

    Returns:
        precision: Scalar `float64` `Tensor` with the value of `true_positives`
            divided by the sum of `true_positives` and `false_positives`.
        update_op: `Operation` that increments `true_positives` and
            `false_positives` variables appropriately, and whose value matches `precision`.

    Raises:
        ValueError: If `weights` is not `None` and its shape doesn't match
            `predictions`, or if either `metrics_collections` or `updates_collections`
            are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - SparsePrecisionAtK: {k: 2}
      # other model properties
    ```
    """
    IDENTIFIER = 'SparsePrecisionAtK'
    SCHEMA = SparsePrecisionAtKSchema

    def __init__(self, k, class_id, **kwargs):
        self.k = k
        self.class_id = class_id
        super(SparsePrecisionAtKConfig, self).__init__(**kwargs)


class MeanAbsoluteErrorSchema(BaseMetricSchema):

    @staticmethod
    def schema_config():
        return MeanAbsoluteErrorConfig


class MeanAbsoluteErrorConfig(BaseMetricConfig):
    """Computes the mean absolute error between the labels and predictions.

    The `streaming_mean_absolute_error` function creates two local variables,
    `total` and `count` that are used to compute the mean absolute error. This
    average is weighted by `weights`, and it is ultimately returned as
    `mean_absolute_error`: an idempotent operation that simply divides `total` by
    `count`.

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the
    `mean_absolute_error`. Internally, an `absolute_errors` operation computes the
    absolute value of the differences between `predictions` and `labels`. Then
    `update_op` increments `total` with the reduced sum of the product of
    `weights` and `absolute_errors`, and it increments `count` with the reduced
    sum of `weights`

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: A `Tensor` of the same shape as `predictions`.
        output_layer: A `Tensor` of arbitrary shape.
        weights: Optional `Tensor` indicating the frequency with which an example is
            sampled. Rank must be 0, or the same rank as `labels`, and must be
            broadcastable to `labels` (i.e., all dimensions must be either `1`, or
            the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that
            `mean_absolute_error` should be added to.
        updates_collections: An optional list of collections that `update_op` should be added to.
        name: An optional variable_scope name.

    Returns:
        mean_absolute_error: A `Tensor` representing the current mean, the value of
            `total` divided by `count`.
        update_op: An operation that increments the `total` and `count` variables
            appropriately and whose value matches `mean_absolute_error`.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - MeanAbsoluteError
      # other model properties
    ```
    """
    IDENTIFIER = 'MeanAbsoluteError'
    SCHEMA = MeanAbsoluteErrorSchema


class MeanRelativeErrorSchema(BaseMetricSchema):
    normalizer = fields.Str()  # name of the normalizer tensor

    @staticmethod
    def schema_config():
        return MeanRelativeErrorConfig


class MeanRelativeErrorConfig(BaseMetricConfig):
    """Computes the mean relative error by normalizing with the given values.

    The `streaming_mean_relative_error` function creates two local variables,
    `total` and `count` that are used to compute the mean relative absolute error.
    This average is weighted by `weights`, and it is ultimately returned as
    `mean_relative_error`: an idempotent operation that simply divides `total` by
    `count`.

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the
    `mean_reative_error`. Internally, a `relative_errors` operation divides the
    absolute value of the differences between `predictions` and `labels` by the
    `normalizer`. Then `update_op` increments `total` with the reduced sum of the
    product of `weights` and `relative_errors`, and it increments `count` with the
    reduced sum of `weights`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: A `Tensor` of the same shape as `predictions`.
        output_layer: A `Tensor` of arbitrary shape.
        normalizer: A `Tensor` of the same shape as `predictions`.
        weights: Optional `Tensor` indicating the frequency with which an example is
            sampled. Rank must be 0, or the same rank as `labels`, and must be
            broadcastable to `labels` (i.e., all dimensions must be either `1`, or
            the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that
            `mean_relative_error` should be added to.
        updates_collections: An optional list of collections that `update_op` should be added to.
        name: An optional variable_scope name.

    Returns:
        mean_relative_error: A `Tensor` representing the current mean, the value of
            `total` divided by `count`.
        update_op: An operation that increments the `total` and `count` variables
            appropriately and whose value matches `mean_relative_error`.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - MeanRelativeError
      # other model properties
    ```
    """
    IDENTIFIER = 'MeanRelativeError'
    SCHEMA = MeanRelativeErrorSchema

    def __init__(self, normalizer, **kwargs):
        self.normalizer = normalizer
        super(MeanRelativeErrorConfig, self).__init__(**kwargs)


class MeanSquaredErrorSchema(BaseMetricSchema):

    @staticmethod
    def schema_config():
        return MeanSquaredErrorConfig


class MeanSquaredErrorConfig(BaseMetricConfig):
    """Computes the mean squared error between the labels and predictions.

    The `streaming_mean_squared_error` function creates two local variables,
    `total` and `count` that are used to compute the mean squared error.
    This average is weighted by `weights`, and it is ultimately returned as
    `mean_squared_error`: an idempotent operation that simply divides `total` by
    `count`.

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the
    `mean_squared_error`. Internally, a `squared_error` operation computes the
    element-wise square of the difference between `predictions` and `labels`. Then
    `update_op` increments `total` with the reduced sum of the product of
    `weights` and `squared_error`, and it increments `count` with the reduced sum
    of `weights`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: A `Tensor` of the same shape as `predictions`.
        output_layer: A `Tensor` of arbitrary shape.
        weights: Optional `Tensor` indicating the frequency with which an example is
            sampled. Rank must be 0, or the same rank as `labels`, and must be
            broadcastable to `labels` (i.e., all dimensions must be either `1`, or
            the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that
            `mean_squared_error` should be added to.
        updates_collections: An optional list of collections that `update_op` should
            be added to.
        name: An optional variable_scope name.

    Returns:
        mean_squared_error: A `Tensor` representing the current mean, the value of
            `total` divided by `count`.
        update_op: An operation that increments the `total` and `count` variables
            appropriately and whose value matches `mean_squared_error`.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - MeanSquaredError
      # other model properties
    ```
    """
    IDENTIFIER = 'MeanSquaredError'
    SCHEMA = MeanSquaredErrorSchema


class RootMeanSquaredErrorSchema(BaseMetricSchema):

    @staticmethod
    def schema_config():
        return RootMeanSquaredErrorConfig


class RootMeanSquaredErrorConfig(BaseMetricConfig):
    """Computes the root mean squared error between the labels and predictions.

    The `streaming_root_mean_squared_error` function creates two local variables,
    `total` and `count` that are used to compute the root mean squared error.
    This average is weighted by `weights`, and it is ultimately returned as
    `root_mean_squared_error`: an idempotent operation that takes the square root
    of the division of `total` by `count`.

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the
    `root_mean_squared_error`. Internally, a `squared_error` operation computes
    the element-wise square of the difference between `predictions` and `labels`.
    Then `update_op` increments `total` with the reduced sum of the product of
    `weights` and `squared_error`, and it increments `count` with the reduced sum
    of `weights`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: A `Tensor` of the same shape as `predictions`.
        output_layer: A `Tensor` of arbitrary shape.
        weights: Optional `Tensor` indicating the frequency with which an example is
            sampled. Rank must be 0, or the same rank as `labels`, and must be
            broadcastable to `labels` (i.e., all dimensions must be either `1`, or
            the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that
            `root_mean_squared_error` should be added to.
        updates_collections: An optional list of collections that `update_op` should be added to.
        name: An optional variable_scope name.

    Returns:
        root_mean_squared_error: A `Tensor` representing the current mean, the value
            of `total` divided by `count`.
        update_op: An operation that increments the `total` and `count` variables
            appropriately and whose value matches `root_mean_squared_error`.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - RootMeanSquaredError
      # other model properties
    ```
    """
    IDENTIFIER = 'RootMeanSquaredError'
    SCHEMA = RootMeanSquaredErrorSchema


class CovarianceSchema(BaseMetricSchema):

    @staticmethod
    def schema_config():
        return CovarianceConfig


class CovarianceConfig(BaseMetricConfig):
    """Computes the unbiased sample covariance between `predictions` and `labels`.

    The `streaming_covariance` function creates four local variables,
    `comoment`, `mean_prediction`, `mean_label`, and `count`, which are used to
    compute the sample covariance between predictions and labels across multiple
    batches of data. The covariance is ultimately returned as an idempotent
    operation that simply divides `comoment` by `count` - 1. We use `count` - 1
    in order to get an unbiased estimate.

    The algorithm used for this online computation is described in
    https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance.
    Specifically, the formula used to combine two sample comoments is
    `C_AB = C_A + C_B + (E[x_A] - E[x_B]) * (E[y_A] - E[y_B]) * n_A * n_B / n_AB`
    The comoment for a single batch of data is simply
    `sum((x - E[x]) * (y - E[y]))`, optionally weighted.

    If `weights` is not None, then it is used to compute weighted comoments,
    means, and count. NOTE: these weights are treated as "frequency weights", as
    opposed to "reliability weights". See discussion of the difference on
    https://wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_variance

    To facilitate the computation of covariance across multiple batches of data,
    the function creates an `update_op` operation, which updates underlying
    variables and returns the updated covariance.

    Args:
        input_layer: A `Tensor` of the same shape as `predictions`.
        output_layer: A `Tensor` of arbitrary shape.
        weights: Optional `Tensor` indicating the frequency with which an example is
            sampled. Rank must be 0, or the same rank as `labels`, and must be
            broadcastable to `labels` (i.e., all dimensions must be either `1`, or
            the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that
            the metric should be added to.
        updates_collections: An optional list of collections that `update_op` should be added to.
        name: An optional variable_scope name.

    Returns:
        covariance: A `Tensor` representing the current unbiased sample covariance,
            `comoment` / (`count` - 1).
        update_op: An operation that updates the local variables appropriately.

    Raises:
        ValueError: If labels and predictions are of different sizes or if either
            `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - RootMeanSquaredError
      # other model properties
    ```
    """
    IDENTIFIER = 'Covariance'
    SCHEMA = CovarianceSchema


class PearsonCorrelationSchema(BaseMetricSchema):

    @staticmethod
    def schema_config():
        return PearsonCorrelationConfig


class PearsonCorrelationConfig(BaseMetricConfig):
    """Computes Pearson correlation coefficient between `predictions`, `labels`.

    The `streaming_pearson_correlation` function delegates to
    `streaming_covariance` the tracking of three [co]variances:

    - `streaming_covariance(predictions, labels)`, i.e. covariance
    - `streaming_covariance(predictions, predictions)`, i.e. variance
    - `streaming_covariance(labels, labels)`, i.e. variance

    The product-moment correlation ultimately returned is an idempotent operation
    `cov(predictions, labels) / sqrt(var(predictions) * var(labels))`. To
    facilitate correlation computation across multiple batches, the function
    groups the `update_op`s of the underlying streaming_covariance and returns an
    `update_op`.

    If `weights` is not None, then it is used to compute a weighted correlation.
    NOTE: these weights are treated as "frequency weights", as opposed to
    "reliability weights". See discussion of the difference on
    https://wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_variance

    Args:
        input_layer: A `Tensor` of the same shape as `predictions`.
        output_layer: A `Tensor` of arbitrary shape.
        weights: Optional `Tensor` indicating the frequency with which an example is
            sampled. Rank must be 0, or the same rank as `labels`, and must be
            broadcastable to `labels` (i.e., all dimensions must be either `1`, or
            the same as the corresponding `labels` dimension).
        metrics_collections: An optional list of collections that
            the metric should be added to.
        updates_collections: An optional list of collections that `update_op` should be added to.
        name: An optional variable_scope name.

    Returns:
        pearson_r: A `Tensor` representing the current Pearson product-moment
            correlation coefficient, the value of
            `cov(predictions, labels) / sqrt(var(predictions) * var(labels))`.
        update_op: An operation that updates the underlying variables appropriately.

    Raises:
        ValueError: If `labels` and `predictions` are of different sizes, or if
            `weights` is the wrong size, or if either `metrics_collections` or
            `updates_collections` are not a `list` or `tuple`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - RootMeanSquaredError
      # other model properties
    ```
    """
    IDENTIFIER = 'PearsonCorrelation'
    SCHEMA = PearsonCorrelationSchema


class MeanCosineDistanceSchema(BaseMetricSchema):
    dim = fields.Int()

    @staticmethod
    def schema_config():
        return MeanCosineDistanceConfig


class MeanCosineDistanceConfig(BaseMetricConfig):
    """Computes the cosine distance between the labels and predictions.

    The `streaming_mean_cosine_distance` function creates two local variables,
    `total` and `count` that are used to compute the average cosine distance
    between `predictions` and `labels`. This average is weighted by `weights`,
    and it is ultimately returned as `mean_distance`, which is an idempotent
    operation that simply divides `total` by `count`.

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the
    `mean_distance`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: A `Tensor` of the same shape as `predictions`.
        output_layer: A `Tensor` of arbitrary shape.
        dim: The dimension along which the cosine distance is computed.
        weights: An optional `Tensor` whose shape is broadcastable to `predictions`,
            and whose dimension `dim` is 1.
        metrics_collections: An optional list of collections that the metric
            value variable should be added to.
        updates_collections: An optional list of collections that the metric update
            ops should be added to.
        name: An optional variable_scope name.

    Returns:
        mean_distance: A `Tensor` representing the current mean, the value of
            `total` divided by `count`.
        update_op: An operation that increments the `total` and `count` variables appropriately.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - MeanCosineDistance
      # other model properties
    ```
    """
    IDENTIFIER = 'MeanCosineDistance'
    SCHEMA = MeanCosineDistanceSchema

    def __init__(self, dim, **kwargs):
        self.dim = dim
        super(MeanCosineDistanceConfig, self).__init__(**kwargs)


class PercentageLessSchema(BaseTensorMetricSchema):
    threshold = fields.Float()

    @staticmethod
    def schema_config():
        return PercentageLessConfig


class PercentageLessConfig(BaseTensorMetricConfig):
    """Computes the percentage of values less than the given threshold.

    The `streaming_percentage_less` function creates two local variables,
    `total` and `count` that are used to compute the percentage of `values` that
    fall below `threshold`. This rate is weighted by `weights`, and it is
    ultimately returned as `percentage` which is an idempotent operation that
    simply divides `total` by `count`.

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the
    `percentage`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        values: A numeric `Tensor` of arbitrary size.
        threshold: A scalar threshold.
        weights: An optional `Tensor` whose shape is broadcastable to `values`.
        metrics_collections: An optional list of collections that the metric
            value variable should be added to.
        updates_collections: An optional list of collections that the metric update
            ops should be added to.
        name: An optional variable_scope name.

    Returns:
        percentage: A `Tensor` representing the current mean, the value of `total`
            divided by `count`.
        update_op: An operation that increments the `total` and `count` variables appropriately.

    Raises:
        ValueError: If `weights` is not `None` and its shape doesn't match `values`,
            or if either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - PercentageLess:
            values: dense_layer_1
            threshold: 1
      # other model properties
    ```
    """
    IDENTIFIER = 'PercentageLess'
    SCHEMA = PercentageLessSchema

    def __init__(self, values, threshold, **kwargs):
        self.threshold = threshold
        super(PercentageLessConfig, self).__init__(values=values, **kwargs)


class MeanIOUSchema(BaseMetricSchema):
    num_classes = fields.Int()

    @staticmethod
    def schema_config():
        return MeanIOUConfig


class MeanIOUConfig(BaseMetricConfig):
    """Calculate per-step mean Intersection-Over-Union (mIOU).

    Mean Intersection-Over-Union is a common evaluation metric for
    semantic image segmentation, which first computes the IOU for each
    semantic class and then computes the average over classes.
    IOU is defined as follows:
      IOU = true_positive / (true_positive + false_positive + false_negative).
    The predictions are accumulated in a confusion matrix, weighted by `weights`,
    and mIOU is then calculated from it.

    For estimation of the metric over a stream of data, the function creates an
    `update_op` operation that updates these variables and returns the `mean_iou`.

    If `weights` is `None`, weights default to 1. Use weights of 0 to mask values.

    Args:
        input_layer: A `Tensor` of ground truth labels with shape [batch size] and of
            type `int32` or `int64`. The tensor will be flattened, if its rank > 1.
        output_layer: A `Tensor` of prediction results for semantic labels, whose
            shape is [batch size] and type `int32` or `int64`. The tensor will be
            flattened, if its rank > 1.
        num_classes: The possible number of labels the prediction task can
            have. This value must be provided, since a confusion matrix of
            dimension = [num_classes, num_classes] will be allocated.
        weights: An optional `Tensor` whose shape is broadcastable to `predictions`.
        metrics_collections: An optional list of collections that `mean_iou` should be added to.
        updates_collections: An optional list of collections `update_op` should be added to.
        name: An optional variable_scope name.

    Returns:
        mean_iou: A `Tensor` representing the mean intersection-over-union.
        update_op: An operation that increments the confusion matrix.

    Raises:
        ValueError: If `predictions` and `labels` have mismatched shapes, or if
            `weights` is not `None` and its shape doesn't match `predictions`, or if
            either `metrics_collections` or `updates_collections` are not a list or tuple.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      metrics:
        - MeanIOU:
            num_classes: 10
      # other model properties
    ```
    """
    IDENTIFIER = 'MeanIOU'
    SCHEMA = MeanIOUSchema

    def __init__(self, num_classes, **kwargs):
        self.num_classes = num_classes
        super(MeanIOUConfig, self).__init__(**kwargs)


class MetricSchema(BaseMultiSchema):
    __multi_schema_name__ = 'Metric'
    __configs__ = {
        TruePositivesConfig.IDENTIFIER: TruePositivesConfig,
        TrueNegativesConfig.IDENTIFIER: TrueNegativesConfig,
        FalsePositivesConfig.IDENTIFIER: FalsePositivesConfig,
        FalseNegativesConfig.IDENTIFIER: FalseNegativesConfig,
        MeanConfig.IDENTIFIER: MeanConfig,
        MeanTensorConfig.IDENTIFIER: MeanTensorConfig,
        AccuracyConfig.IDENTIFIER: AccuracyConfig,
        PrecisionConfig.IDENTIFIER: PrecisionConfig,
        RecallConfig.IDENTIFIER: RecallConfig,
        AUCConfig.IDENTIFIER: AUCConfig,
        SpecificityAtSensitivityConfig.IDENTIFIER: SpecificityAtSensitivityConfig,
        SensitivityAtSpecificityConfig.IDENTIFIER: SensitivityAtSpecificityConfig,
        PrecisionAtThresholdsConfig.IDENTIFIER: PrecisionAtThresholdsConfig,
        RecallAtThresholdsConfig.IDENTIFIER: RecallAtThresholdsConfig,
        SparseRecallAtKConfig.IDENTIFIER: SparseRecallAtKConfig,
        SparsePrecisionAtKConfig.IDENTIFIER: SparsePrecisionAtKConfig,
        MeanAbsoluteErrorConfig.IDENTIFIER: MeanAbsoluteErrorConfig,
        MeanRelativeErrorConfig.IDENTIFIER: MeanRelativeErrorConfig,
        MeanSquaredErrorConfig.IDENTIFIER: MeanSquaredErrorConfig,
        RootMeanSquaredErrorConfig.IDENTIFIER: RootMeanSquaredErrorConfig,
        CovarianceConfig.IDENTIFIER: CovarianceConfig,
        PearsonCorrelationConfig.IDENTIFIER: PearsonCorrelationConfig,
        MeanCosineDistanceConfig.IDENTIFIER: MeanCosineDistanceConfig,
        PercentageLessConfig.IDENTIFIER: PercentageLessConfig,
        MeanIOUConfig.IDENTIFIER: MeanIOUConfig,
    }
